@@ -17,6 +17,27 @@ import pysodium
 from multidict import MultiDict  # base class for mdict defined below
 from orderedset import OrderedSet as oset
 
+# Utilities
+def isign(i):
+    """
+    Returns 1 if i > 0, -1 if i < 0, 0 otherwise
+    Integer sign function
+    """
+    return (1 if i > 0 else -1 if i < 0 else 0)
+
+
+def sceil(r):
+    """
+    Returns integer value that is symmetric ceiling of r away from zero
+    Symmetric ceiling function
+    Because int() provides a symmetric floor towards zero, just inc int(r) by:
+     1 when r - int(r) >  0  (r positive)
+    -1 when r - int(r) <  0  (r negative)
+     0 when r - int(r) == 0  (r integral already)
+    abs(r) > abs(int(r) or 0 when abs(r)
+    """
+    return (int(r) + isign(r - int(r)))
+
 
 def datify(cls, d):
     """
@@ -26,8 +47,8 @@ def datify(cls, d):
     d is dict
     """
     try:
-        fieldtypes = {f.name:f.type for f in dataclasses.fields(cls)}
-        return cls(**{f:datify(fieldtypes[f], d[f]) for f in d})  # recursive
+        fieldtypes = {f.name: f.type for f in dataclasses.fields(cls)}
+        return cls(**{f: datify(fieldtypes[f], d[f]) for f in d})  # recursive
     except:
         return d  # Not a dataclass field
 
@@ -75,6 +96,7 @@ def verify64uEd25519(signature, message, verkey):
     msg = message.encode("utf-8")
     return (verifyEd25519(sig, msg, vk))
 
+
 class mdict(MultiDict):
     """
     Multiple valued dictionary. Insertion order of keys preserved.
@@ -189,6 +211,7 @@ def nonStringIterable(obj):
     """
     return (not isinstance(obj, (str, bytes)) and isinstance(obj, Iterable))
 
+
 def nonStringSequence(obj):
     """
     Returns True if obj is non-string sequence, False otherwise
@@ -287,7 +310,7 @@ def toIso8601(dt=None):
 
 def fromIso8601(dts):
     """
-    Returns datetime object from ISO 8601 formated str
+    Returns datetime object from ISO 8601 formated str or bytes
     Converts dts from ISO 8601 format to datetime object
 
     YYYY-MM-DDTHH:MM:SS.ffffff+HH:MM[:SS[.ffffff]]
@@ -296,4 +319,6 @@ def fromIso8601(dts):
     Assumes TZ aware
     For nanosecond use instead attotime or datatime64 in pandas or numpy
     """
+    if hasattr(dts, "decode"):
+        dts = dts.decode("utf-8")
     return (datetime.datetime.fromisoformat(dts))
